long int random(void);
#include "main.h"

typedef struct frontier2 {
  int x, y;
  int weight;
  struct frontier2 *next;
} frontier2;

frontier2 *insert(frontier2 *f, int x, int y, int w)
{
  frontier2 *g, *h;

  h = (frontier2*)calloc(1, sizeof(frontier2));
  h->x = x;
  h->y = y;
  h->weight = w;
  h->next   = NULL;

  if(!f)
    return h;
  if(!f->next) 
    {
      if(f->weight > w) 
	{
	  h->next = f;
	  return h;
	} else {
	  f->next = h;
	  return f;
	}
    }

  for(g = f; g && g->next; g=g->next)
    if((g->weight < w) && (g->next->weight >= w))
	break;

  h->next = g->next;
  g->next = h;
  return f;
}

static void bellmann_ford(int **weights, unsigned int **dist, board *b, int xstart, int ystart)
{
  frontier2 *f, *nf;

  f = (struct frontier2*)calloc(1, sizeof(frontier2));
  f->x = xstart;
  f->y = ystart;
  f->weight = 0;
  f->next   = NULL;

  do {
    nf = f->next;
    if( (dist[f->x][f->y] == (unsigned int)-1) &&
	((b->tiles[f->x][f->y].terrain == PLAIN) ||
	 (b->tiles[f->x][f->y].terrain == BASE) ) )
      {
	dist[f->x][f->y] = f->weight;
	if(dist[f->x+1][f->y] == (unsigned int)(-1))
	  nf = insert(nf, f->x+1, f->y, weights[f->x+1][f->y]+f->weight+1);
	if(dist[f->x-1][f->y] == (unsigned int)(-1))
	  nf = insert(nf, f->x-1, f->y, weights[f->x-1][f->y]+f->weight+1);
	if(dist[f->x][f->y+1] == (unsigned int)(-1))
	  nf = insert(nf, f->x, f->y+1, weights[f->x][f->y+1]+f->weight+1);
	if(dist[f->x][f->y-1] == (unsigned int)(-1))
	  nf = insert(nf, f->x, f->y-1, weights[f->x][f->y-1]+f->weight+1);
      }
    free(f);
    f = nf;
  } while(f);

}

#define min(a, b) ((a)<(b)?(a):(b))

char bestmove(env *E, bedford_entry *target)
{
  unsigned int m;
  unsigned int n,s,e,w;
  static char tab[] = "NSEWNSWENESWNEWSNWESNWSESNEWSNWESEWNSENWSWENSWNEESWNESNWENWSENSWEWNSEWSNWSNEWSENWNSEWNESWENSWESN";
  char *t;

  n=target->dist[E->p->x][E->p->y+1];
  s=target->dist[E->p->x][E->p->y-1];
  e=target->dist[E->p->x+1][E->p->y];
  w=target->dist[E->p->x-1][E->p->y];

  m = min(min(n, s), min(e, w));
  t = tab+((random() % 24) * 4);
  while(*t) {
    if((*t == NORTH) && (n == m))
      return NORTH;
    if((*t == SOUTH) && (s == m))
      return SOUTH;
    if((*t == EAST) && (e == m))
      return EAST;
    if((*t == WEST) && (w == m))
      return WEST;
    t++;
  }
  assert( 0 );
  return 0;
}

bedford_table *create_bedford_table(int maxsize)
{
  bedford_table *bt = (bedford_table *)calloc(1, sizeof(bedford_table));
  printf("creating bedford database for at most %d tables\n",  maxsize);
  bt->maxsize = maxsize;
  return bt;
}

static bedford_entry *lookup_bedford_table(bedford_table *bt, int x, int y)
{
  bedford_entry *t;
  for(t = bt->first; t; t=t->next)
    if( (t->targetx == x) && (t->targety == y) )
      return t;
  return NULL;
}

bedford_entry *bedford_lookup_target(env *e, int x, int y)
{
  bedford_entry *be, *le;
  int X, Y;

  be = lookup_bedford_table(e->bt, x, y);
  if(be)
    {
      if(be->prev != NULL) /* not already the head */
	{
	  if(be->next)     /* not tail */
	    be->next->prev = be->prev;
	  else
	    e->bt->last = be->prev;
	  be->prev->next = be->next;
	  be->next = e->bt->first;
	  be->next->prev = be;
	  e->bt->first = be;
	  be->prev=NULL;
	}
    }
  else
    {
      be = (bedford_entry *)calloc(1, sizeof(bedford_entry));
      be->targetx = x;
      be->targety = y;
      be->dist    = (unsigned int **)malloc(e->b->width*sizeof(unsigned int *));
      for(X = 0; X < e->b->width; X++) {
	be->dist[X] = (unsigned int *)malloc(e->b->height*sizeof(unsigned int));
	for(Y = 0; Y < e->b->height; Y++)
	  be->dist[X][Y] = (unsigned int)(-1);
      }

      bellmann_ford(e->topo_layer, be->dist, e->b, x, y);

      be->next = e->bt->first;
      if(e->bt->first)
	e->bt->first->prev = be;
      e->bt->first = be;
      if(e->bt->cursize == e->bt->maxsize) {
	le = e->bt->last;
	assert( le != NULL );
	le->prev->next = NULL;
	e->bt->last = le->prev;
	for(X = 0; X < e->b->width; X++)
	  free(le->dist[X]);
	free(le->dist);
	free(le);
      } else {
	if(e->bt->last == NULL)
	  e->bt->last = be;
	e->bt->cursize++;
      }
    }

  return be;
}

