#include "main.h"

env *init(char *c_server, unsigned short port)
{
  env *e = (env*)calloc(1, sizeof(env));
  char *initpos;
  
  e->state = EMPTY;

  e->soc = init_net(c_server, port);
  if(e->soc < 0)
    return NULL;
  
  send(e->soc, "Player\n",7,0);
#ifdef DEBUG_NET
  printf("C: Player\n");
#endif

  e->b = get_board(e->soc);

  gui_write( &e->b->width, sizeof(int));
  gui_write( &e->b->height, sizeof(int));

  if (!e->b)
    {
      printf( "could not get board\n");
      return NULL;
    }
  e->bs=init_bases(e);
  init_topo_layer(e);

  e->p = get_good_player(e->soc);
  if(!e->p)
    {
      printf( "could not get player initial configuration\n");
      return NULL;
    }

  e->initial_capacity = e->p->capacity;
  e->initial_money    = e->p->money;

  initpos = getnextline(e->soc);
  if(!initpos)
    {
      printf( "could not inital players position\n");
      return NULL;
    }
  if( parse_server_response(initpos, e) < 0)
    {
      printf( "could not parse initial players position\n");
      return NULL;
    }

  e->bt = create_bedford_table( MAX_MEMORY_SPACE_FOR_BEDFORD_TABLES_IN_BYTES / (sizeof(int) * e->b->width * e->b->height ) );

  return e;
}

void place_players_on_board(env *e)
{
  player *p;
  int x, y;

  for(x = 0; x < e->b->width; x++)
    for(y = 0; y < e->b->height; y++)
      e->b->tiles[x][y].id = -1;

  p = e->p;
  if( (p->x >= 1) && (p->y >= 1) && (!p->dead))
    e->b->tiles[p->x][p->y].id = p->id;
  for(p = e->adv.first; p; p=p->next)
    if( (p->x >= 1) && (p->y >= 1) && (!p->dead) )
    e->b->tiles[p->x][p->y].id = p->id;
}

static void mark_all_player_dead(env *e)
{
  player *p;
  e->p->dead = 1;
  for(p = e->adv.first; p; p = p->next)
    p->dead = 1;
}

void display_env(FILE *stream, env *e)
{ 
#ifdef GUI
  int cp = 0;
  player *p;
  pkt_list_entry *pk;

  display_board(stream, e);
  display_player(stream, e->p);
  for(p = e->adv.first; p; p = p->next) cp++;
  gui_write(&cp, sizeof(int));
  for(p = e->adv.first; p; p = p->next)
    display_player(stream, p);
  display_packages(stream, &e->known);
  
  cp = 0;
  for(pk = e->p->packages.first; pk; pk=pk->next) cp++;
  gui_write(&cp, sizeof(int));
  for(pk = e->p->packages.first; pk; pk=pk->next)
    gui_write(&pk->entry->id, sizeof(int));
#endif
}

int parse_server_response(char *response, env *e)
{
  char *r = response;
  int id, id2;
  player *p;
  base *b;
  pkt_entry *pe;

  while(*r == '#')
    {
      r++;
      id = strtol(r, &r, 10);	      
      p = player_get_or_create_by_id(id, e);
      p->dead = 0;
      assert(p != NULL);
      if(*r == ' ')
	r++;
      while( (*r != '#') && (*r != 0) ) 
	{
	  switch(*r)
	    {
	    case 'X':
	      r++;
	      assert( *r == ' ');
	      r++;
	      p->x = strtol(r, &r, 10);
	      if( (p->x != -1) && (p->y != -1) )
		player_update_packages_position(p);
	      break;
	    case 'Y':
	      r++;
	      assert( *r == ' ');
	      r++;
	      p->y = strtol(r, &r, 10);
	      if( (p->x != -1) && (p->y != -1) )
		player_update_packages_position(p);
	      break;
	    case 'P':
	      r++;
	      assert( *r == ' ');
	      r++;
	      id2 = strtol(r, &r, 10);
	      pe = pkt_lookup_or_add_by_id(e, id2);
	      pe->x = p->x;
	      pe->y = p->y;
	      pe->state=PCARRIED;
	      if(pe->weight != -1)
		p->capacity -= pe->weight;
	      pkt_add_entry(&p->packages, pe);
	      b = find_base(e->bs, pe->x, pe->y);
	      if(b && b->packages)
		{
		  pkt_remove_entry(b->packages, pe);
		  base_update_limit(b);
		}
	      break;
	    case 'D':
	      r++;
	      assert( *r == ' ');
	      r++;
	      id2 = strtol(r, &r, 10);
	      pe = pkt_lookup_or_add_by_id(e, id2);
	      pe->x = p->x;
	      pe->y = p->y;
	      if( (pe->x == pe->xhome) &&
		  (pe->y == pe->yhome) ) 
		{
		  if( p == e->p )
		    pe->state = PDESTROYEDBYUS;
		  else
		    pe->state = PDESTROYED;
		}
	      else
		{
		  pe->state = PONFLOOR;
		}
	      pkt_remove_entry(&p->packages, pe);
	      b = find_base(e->bs, pe->x, pe->y);
	      if(b)
		{
		  if(!b->packages) b->packages = pkt_init();
		  pkt_add_entry(b->packages, pe);
		  base_update_limit(b);
		}
	      if( (pe->weight != -1) )
		p->capacity += pe->weight;
	      break;
	    case 'E':
	      r++;
	      p->x++;
	      if( (p->x != -1) && (p->y != -1) )
		player_update_packages_position(p);
	      break;
	    case 'N':
	      r++;
	      p->y++;
	      if( (p->x != -1) && (p->y != -1) )
		player_update_packages_position(p);
	      break;
	    case 'W':
	      r++;
	      p->x--;
	      if( (p->x != -1) && (p->y != -1) )
		player_update_packages_position(p);
	      break;
	    case 'S':
	      r++;
	      p->y--;
	      if( (p->x != -1) && (p->y != -1) )
		player_update_packages_position(p);
	      break;
	    default:
	      printf( "Unknown protocol operator %c\n Response = %s\n Parsing  : %s\n", *r, response, r);
	      return -1;
	    }
	  if( *r == 0)
	    break;
	  assert(*r == ' ');
	  r++;
	}
    }

  return 0;
}

int round(env *e)
{
  char *newpos;
  char *pklist;

  pklist = getnextline(e->soc);
  if(!pklist)
    {
      printf( "Network error while receiving pkg list\n");
      return -1;
    }
  if (e->p->available)
    pkt_free(e->p->available);

  if( (e->p->available = parse_package_list(pklist, e)) == NULL)
    {
      printf( "could not parse pkg list\n");
      return -1;
    }

  place_players_on_board(e);

  decision(e);

  newpos = getnextline(e->soc);
  if(!newpos)
    {
      printf( "Network error while receiving players position\n");
      return -1;
    }
  mark_all_player_dead(e);
  if( parse_server_response(newpos, e) < 0)
    {
      printf( "could not parse new players position\n");
      return -1;
    }

  return e->p->money;
}

void move(int bid, env *e, char direction)
{
  char order[1024]; /* TODO check */
  /*printf("moving\n");*/
  sprintf(order, "%d Move %c\n", bid, direction);
  send(e->soc, order, strlen(order), 0);
#ifdef DEBUG_NET
  printf( "C: %s", order);
#endif
  e->p->money -= vabs(bid);
}

void stay(int bid,env *e)
{
  char order[1024]; 
  sprintf(order, "%d Pick\n", bid);
  send(e->soc, order, strlen(order), 0);
#ifdef DEBUG_NET
  printf( "C: %s", order);
#endif
  e->p->money -= vabs(bid);
}

void pick(int bid, env *e, pkt_list *pl)
{
  static char *order = NULL;
  static int ordersize = 0;
  char value[64]; /* TODO Check */
  pkt_list_entry *ple;

  if(!order)
    order = (char*)calloc(ordersize = 1024, sizeof(char));
  sprintf(order, "%d Pick ", bid);
  for(ple = pl->first; ple; ple = ple->next) 
    {
      if(ple->next)
	sprintf(value, "%d ", ple->entry->id);
      else
	sprintf(value, "%d", ple->entry->id);

      if( ordersize < strlen(order) + strlen(value) + 6 )
	order = (char*)realloc(order, ordersize = ordersize * 2);

      strcat(order, value);
    }
  strcat(order, "\n");

  send(e->soc, order, strlen(order), 0);
#ifdef DEBUG_NET
  printf( "C: %s", order);
#endif
  e->p->money -= vabs(bid);
}

void drop(int bid, env *e, pkt_list *pl)
{
  static char *order = NULL;
  static int ordersize = 0;
  char value[64]; /* TODO Check */
  pkt_list_entry *ple;

  if(!order)
    order = (char*)calloc(ordersize = 1024, sizeof(char));
  sprintf(order, "%d Drop ", bid);
  for(ple = pl->first; ple; ple = ple->next) 
    {
      if(ple->next)
	sprintf(value, "%d ", ple->entry->id);
      else
	sprintf(value, "%d", ple->entry->id);

      if( ordersize < strlen(order) + strlen(value) + 6 )
	order = (char*)realloc(order, ordersize = ordersize * 2);

      strcat(order, value);
    }
  strcat(order, "\n");

  send(e->soc, order, strlen(order), 0);
#ifdef DEBUG_NET
  printf( "C: %s", order);
#endif
  e->p->money -= vabs(bid);
}
