#include <gdk/gdk.h>
#include <gtk/gtk.h>

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/signal.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

GtkWidget *window, *darea, *sbar, *myframe, *pframe, *pkframe, *packages;
GtkWidget *hbox2, *vbox2, *myid, *mymoney, *mycapacity, *pvbox;
GtkWidget *vbox, *hbox, *label, *sw, *mypackets;
GIOChannel *gsock;
int contextid;
int sock;

#define TILESIZE 6

int width, height;
int WIDTH, HEIGHT;
guchar *rgbbuf;

struct bid {
  char type;
  int id;
  unsigned int distance;
};

#define Red(x, y)   rgbbuf[ (y) * WIDTH * 3 + (x)*3 + 0 ]
#define Green(x, y) rgbbuf[ (y) * WIDTH * 3 + (x)*3 + 1 ]
#define Blue(x, y)  rgbbuf[ (y) * WIDTH * 3 + (x)*3 + 2 ]

struct rgbbyid {
  int id;
  guchar r, g, b;
  struct rgbbyid *next;
  GtkWidget *label;
  GtkWidget *eb;
  char alive;
} *IDs = NULL;

struct colordesc {
  int max;
  unsigned char r, g, b;
};

struct package {
  int id;
  int w;
  int x, y;
  int hx, hy;
  char d;
  struct package *next;
};

struct colordesc colors[] = { {0,   255, 255, 255},
			      {128,   0, 255,   0},
			      {256,   0, 255, 255},
			      {372, 128, 255, 255},
                              {-1,    0,   0,   0}};
unsigned int maxcolors;
unsigned int *cmap;

void buildcmap()
{
  int i, j;
  int min, max;
  int rm, rM, gm, gM, bm, bM;
  double decr, decg, decb;
  int nbdesc;

  for(nbdesc = 0; colors[nbdesc].max != -1; nbdesc++);

  maxcolors = colors[nbdesc-1].max;
  cmap = (unsigned int *)calloc(maxcolors, sizeof(unsigned int));

  for(i = 1; i < nbdesc; i++) {
    min = colors[i-1].max;
    max = colors[i].max-1;
    rm  = colors[i-1].r;
    rM  = colors[i].r;
    gm  = colors[i-1].g;
    gM  = colors[i].g;
    bm  = colors[i-1].b;
    bM  = colors[i].b;
    decr = (double)(rM-rm)/(double)(max-min);
    decg = (double)(gM-gm)/(double)(max-min);
    decb = (double)(bM-bm)/(double)(max-min);
    printf("From %d %02x%02x%02x to %d %02x%02x%02x\n", min, rm, gm, bm,
	   max, rM, gM, bM);
    for(j = min; j < max; j++) {
      cmap[j] = (unsigned int)
	( (((int)(rm + (double)(j-min)*decr)) << 16) +
	  (((int)(gm + (double)(j-min)*decg)) << 8 ) +
	  (((int)(bm + (double)(j-min)*decb)) << 0 ) );
    }
  }
}

void settile_player(int x, int y, int id)
{
  int i, j;
  static int first = 1;
  struct rgbbyid *p;
  char value[128];
   
  for(p = IDs; p; p=p->next)
    if(p->id == id)
      break;
  if(!p)
    {
      GtkRcStyle *rc_style;
      GdkColor color;

      p = (struct rgbbyid *)calloc(1, sizeof(struct rgbbyid));
      p->id = id;
      if( first ) 
	{
	  p->r = 255;
	  p->g = 0;
	  p->b = 0;
	  first = 0;
	} else {
	  p->r = random() % 256;
	  p->g = random() % 256;
	  p->b = random() % 256;
	}
      p->next = IDs;
      p->alive = 1;
      sprintf(value, "player %d (alive)", id);
      p->eb = gtk_event_box_new();
      p->label = gtk_label_new(value);

      color.red   = p->r * (65535/255);
      color.green = p->g * (65535/255);
      color.blue  = p->b * (65535/255);
      color.pixel = (gulong)(p->r*65536 + p->g*256 + p->b);
      gdk_color_alloc(gtk_widget_get_colormap(p->label), &color);
      
      rc_style = gtk_rc_style_new();
      rc_style->fg[GTK_STATE_NORMAL] = color;
      rc_style->color_flags[GTK_STATE_NORMAL] |= GTK_RC_FG;
      gtk_widget_modify_style (p->eb, rc_style);
      gtk_widget_modify_style (p->label, rc_style);
      gtk_rc_style_unref (rc_style);

      gtk_widget_hide_all(pvbox);
      gtk_widget_show(p->eb);
      gtk_box_pack_start(GTK_BOX(pvbox), p->eb, FALSE, FALSE, 1);
      gtk_widget_show(p->label);
      gtk_container_add(GTK_CONTAINER(p->eb), p->label);
      gtk_widget_show_all(pvbox);

      IDs = p;
    }

  for(i = 0; i < TILESIZE; i++) {
    Red(x*TILESIZE+i, y*TILESIZE)   = p->r;
    Green(x*TILESIZE+i, y*TILESIZE) = p->g;
    Blue(x*TILESIZE+i, y*TILESIZE)  = p->b;    
    Red(x*TILESIZE+i, (y+1)*TILESIZE-1)   = p->r;
    Green(x*TILESIZE+i, (y+1)*TILESIZE-1) = p->g;
    Blue(x*TILESIZE+i, (y+1)*TILESIZE-1)  = p->b;
  }
  for(j = 0; j < TILESIZE; j++) {
    Red(x*TILESIZE, y*TILESIZE+j)   = p->r;
    Green(x*TILESIZE, y*TILESIZE+j) = p->g;
    Blue(x*TILESIZE, y*TILESIZE+j)  = p->b;    
    Red((x+1)*TILESIZE-1, y*TILESIZE+j)   = p->r;
    Green((x+1)*TILESIZE-1, y*TILESIZE+j) = p->g;
    Blue((x+1)*TILESIZE-1, y*TILESIZE+j)  = p->b;
    }
}

void settile_base(int x, int y)
{
  int i, j;
  for(i = 1; i < TILESIZE-1; i++) 
    for(j = 1; j < TILESIZE-1; j++)
      {
	Red(x*TILESIZE+i, y*TILESIZE+j)   = 255;
	Green(x*TILESIZE+i, y*TILESIZE+j) = 0;
	Blue(x*TILESIZE+i, y*TILESIZE+j)  = 0;
      }
}

void settile_water(int x, int y)
{
  int i, j;
  for(i = 0; i < TILESIZE; i++)
    for(j = 0; j < TILESIZE; j++) {
      Red(x*TILESIZE+i, y*TILESIZE+j)   = 0;
      Green(x*TILESIZE+i, y*TILESIZE+j) = 0;
      Blue(x*TILESIZE+i, y*TILESIZE+j)  = 255;
    }
}

void settile_wall(int x, int y)
{
  int i, j;
  for(i = 0; i < TILESIZE; i++)
    for(j = 0; j < TILESIZE; j++) {
      Red(x*TILESIZE+i, y*TILESIZE+j)   = 0;
      Green(x*TILESIZE+i, y*TILESIZE+j) = 0;
      Blue(x*TILESIZE+i, y*TILESIZE+j)  = 0;
    }
}

void settile_plain(int x, int y, unsigned int distance)
{
  int i, j;
  for(i = 0; i < TILESIZE; i++)
    for(j = 0; j < TILESIZE; j++) {
      if(distance < maxcolors) {
	Red(x*TILESIZE+i, y*TILESIZE+j)   = (cmap[distance] & 0xff0000)>> 16;
	Green(x*TILESIZE+i, y*TILESIZE+j) = (cmap[distance] & 0x00ff00)>>  8;
	Blue(x*TILESIZE+i, y*TILESIZE+j)  = (cmap[distance] & 0x0000ff)>>  0;
      } else {
	Red(x*TILESIZE+i, y*TILESIZE+j)   = 255;
	Green(x*TILESIZE+i, y*TILESIZE+j) = 255;
	Blue(x*TILESIZE+i, y*TILESIZE+j)  = 255;
      }
    }
}

gboolean on_focus_out ( GtkWidget *widget, GdkEventFocus *event,
			gpointer user_data)
{
  gtk_statusbar_pop(GTK_STATUSBAR(sbar), contextid);
  return TRUE;
}

gboolean on_mouse_move( GtkWidget *widget, GdkEventMotion *event,
			gpointer user_data)
{
  static int X = -1, Y = -1;
  char msg[1024];

  if( (X != (event->x / TILESIZE)) ||
      (Y != height - (event->y / TILESIZE)) )
    {
      gtk_statusbar_pop(GTK_STATUSBAR(sbar), contextid);
      X = event->x / TILESIZE;
      Y = height - (event->y / TILESIZE);
      sprintf(msg, "Coordinates on map : %d, %d", X, Y);
      gtk_statusbar_push(GTK_STATUSBAR(sbar), contextid, msg);
    }
  return TRUE;
}

gboolean on_darea_expose (GtkWidget *widget,
			  GdkEventExpose *event,
			  gpointer user_data)
{
  gdk_draw_rgb_image (widget->window, widget->style->fg_gc[GTK_STATE_NORMAL],
		      0, 0, WIDTH, HEIGHT,
		      GDK_RGB_DITHER_MAX, rgbbuf, WIDTH * 3);
  return FALSE;
}

void read_player(int *id, int *money, int *capacity, char *dead)
{
  read(sock, id, sizeof(int));
  read(sock, money, sizeof(int));
  read(sock, capacity, sizeof(int));
  read(sock, dead, sizeof(char));
}

static struct package *AllPackages = NULL;

struct package *get_package_from_id(int id)
{
  struct package *p;
  for(p = AllPackages; p; p = p->next)
    if(p->id == id) 
      return p;
  return NULL;
}

struct package *update_packages(int id, int w, int x, int y, int hx, 
				int hy, char d)
{
  struct package *p;
  for(p = AllPackages; p; p = p->next)
    if(p->id == id) 
      {
	if( (p->w != w) || (p->x != x) || (p->y != y) || (p->hx != hx)
	    || (p->hy != hy) || (p->d != d) )
	  {
	    p->w = w;
	    p->x = x;
	    p->y = y;
	    p->hx = hx;
	    p->hy = hy;
	    p->d = d;
	    return p;
	  }
	return NULL;
      }
  p = (struct package *)calloc(1, sizeof(struct package));
  p->id = id;
  p->w  = w;
  p->x  = x;
  p->y  = y;
  p->hx = hx;
  p->hy = hy;
  p->d  = d;
  p->next = AllPackages;
  AllPackages = p;
  return p;
}

gboolean read_state()
{
  int x, y;
  GdkRectangle r;
  struct bid b;
  int id, money, capacity;
  struct package *pack;
  int nb, i, total, w, hx, hy;
  char value[64];
  char **values;
  double vpos;
  char dead;
  struct rgbbyid *rid;
  int row;
  int frozen;
  char d;
  char *PSTATE[] = { "_", "x", "^", "?", "!" };

  for(y = height-1; y >= 0; y--)
    for(x = 0; x < width; x++)
      {
	if( read(sock, &b, sizeof(struct bid)) != sizeof(struct bid) )
	  {
	    perror("read");
	    return FALSE;
	  }
	switch(b.type) {
	case '#' : settile_wall(x, y); break;
	case '~' : settile_water(x, y); break;
	case '.' : settile_plain(x, y, b.distance); break;
	case '@' : 
	  settile_plain(x, y, b.distance);
	  settile_base(x, y);
	  break;
	}
	if(b.id != -1)
	  settile_player(x, y, b.id);
      }

  read_player(&id, &money, &capacity, &dead);
  sprintf(value, "%d", id);
  gtk_label_set_text(GTK_LABEL(myid), value);
  sprintf(value, "%d", money);
  gtk_label_set_text(GTK_LABEL(mymoney), value);
  sprintf(value, "%d", capacity);
  gtk_label_set_text(GTK_LABEL(mycapacity), value);

  read(sock, &nb, sizeof(int));
  for(i = 0; i < nb; i++) {
    read_player(&id, &money, &capacity, &dead);
    for(rid = IDs; rid; rid=rid->next)
      if(rid->id == id) {
	if(dead && rid->alive) {
	  rid->alive = 0;
	  sprintf(value, "player %d (dead)", rid->id);
	  gtk_label_set_text(GTK_LABEL(rid->label), value);
	}
	if(!dead && !rid->alive) {
	  rid->alive = 1;
	  sprintf(value, "player %d (alive)", rid->id);
	  gtk_label_set_text(GTK_LABEL(rid->label), value);
	}
	break;
      }
  }

  read(sock, &nb, sizeof(int));
  read(sock, &total, sizeof(int));
  values = (char**)malloc(4*sizeof(char*));
  for(i = 0; i < 4; i++)
    values[i] = (char*)malloc(128);
  vpos = gtk_clist_get_vadjustment(GTK_CLIST(packages))->value;
  for(i = 0; i < nb; i++) {
    read(sock, &id, sizeof(int));
    read(sock, &w, sizeof(int));
    read(sock, &x, sizeof(int));
    read(sock, &y, sizeof(int));
    read(sock, &hx, sizeof(int));
    read(sock, &hy, sizeof(int));
    read(sock, &d, sizeof(char));
    if( (pack = update_packages(id, w, x, y, hx, hy, d)) ) 
      {	
	if( !frozen )
	  {
	    gtk_clist_freeze(GTK_CLIST(packages));
	    frozen = 1;
	  }
	sprintf(values[0], "%s %d", PSTATE[(unsigned int)d-1], id);
	sprintf(values[1], "%d, %d", x, y);
	sprintf(values[2], "%d", w);
	sprintf(values[3], "%d, %d", hx, hy);
	if( (row = gtk_clist_find_row_from_data(GTK_CLIST(packages), pack)) 
	    < 0)
	  row = gtk_clist_append(GTK_CLIST(packages), values);
	else 
	  {
	    gtk_clist_remove(GTK_CLIST(packages), row);
	    row = gtk_clist_insert(GTK_CLIST(packages), row, values);
	  }
	gtk_clist_set_row_data(GTK_CLIST(packages), row, pack);
       }
  }
  if(frozen)
    {
      gtk_clist_thaw(GTK_CLIST(packages));
      frozen = 0; 
      
    }
  gtk_adjustment_set_value(gtk_clist_get_vadjustment(GTK_CLIST(packages)), 
			   vpos);

  read(sock, &nb, sizeof(int));
  gtk_clist_freeze(GTK_CLIST(mypackets));
  gtk_clist_clear(GTK_CLIST(mypackets));
  for(i = 0; i < nb; i++) {
    read(sock, &id, sizeof(int));
    pack = get_package_from_id(id);
    sprintf(values[0], "%s %d", PSTATE[(unsigned int)(pack->d)-1], id);
    sprintf(values[1], "%d", pack->w);
    sprintf(values[2], "%d, %d", pack->hx, pack->hy);
    gtk_clist_append(GTK_CLIST(mypackets), values);
  }
  gtk_clist_thaw(GTK_CLIST(mypackets));

  r.x = 0; r.y = 0; r.width = WIDTH; r.height = HEIGHT;
  gtk_widget_draw(GTK_WIDGET(darea), &r);

  for(i = 0; i < 4; i++) {
    free(values[i]);
    values[i] = NULL;
  }
  free(values);

  return TRUE;
}

gboolean on_data_in(GIOChannel *source,
		    GIOCondition condition,
		    gpointer data)
{
  read_state();
  return read_state();
}

void onchld(int s)
{
  wait(NULL);
  signal(SIGCHLD, onchld);
}

int main(int argc, char *argv[])
{
  struct sockaddr_in remote;
  int listener;
  int rlen;
  int size[2];
  static char *packagestitle[] = { "Id                 ", 
				   "Location", "Weight", "Home" };
  static char *mypacketstitles[] = { "Id                ", 
				     "Weight", "Home" };

  if(argc < 2)
    {
      fprintf(stderr, "Usage : %s [listening port]\n", argv[0]);
      exit(1);
    }

  buildcmap();

  listener = socket(AF_INET, SOCK_STREAM, 0);
  remote.sin_addr.s_addr = INADDR_ANY;
  remote.sin_port = htons(atoi(argv[1]));
  remote.sin_family = AF_INET;
  if(bind(listener, (struct sockaddr*)&remote, sizeof(remote)) < 0 )
    perror("bind"), exit(0);
  if(listen(listener, 1) < 0)
    perror("bind"), exit(0);
  rlen = sizeof(remote);
  
  signal(SIGCHLD, onchld);

  while(1) 
    {
      sock = accept(listener, (struct sockaddr*)&remote, &rlen);
      if(sock < 0)
	perror("accept");
      else {
	if(fork()==0) {
	  close(listener);
	  
	  gtk_init(&argc, &argv);
	  gdk_rgb_init();

	  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	  gtk_signal_connect(GTK_OBJECT(window), "destroy",
			     GTK_SIGNAL_FUNC(gtk_main_quit), NULL);

	  vbox = gtk_vbox_new(FALSE, 0);
	  gtk_container_add(GTK_CONTAINER(window), vbox);
	  
	  hbox = gtk_hbox_new(FALSE, 0);
	  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);
	  
	  sbar = gtk_statusbar_new();
	  contextid = gtk_statusbar_get_context_id(GTK_STATUSBAR(sbar),
						   "Location");
	  gtk_box_pack_start(GTK_BOX(vbox), sbar, FALSE, FALSE, 0);
	  
	  read(sock, size, 2*sizeof(int));
      
	  darea = gtk_drawing_area_new();
	  width = size[0];
	  height = size[1];
	  WIDTH = TILESIZE*width;
	  HEIGHT = TILESIZE*height;
	  rgbbuf  = (guchar *)calloc( 3 * WIDTH * HEIGHT, 1 );
	  memset(rgbbuf, 255,  3 * WIDTH * HEIGHT);
	  printf("Setting area size to %dx%d\n", WIDTH, HEIGHT);
	  gtk_drawing_area_size(GTK_DRAWING_AREA(darea), WIDTH, HEIGHT);
	  gtk_box_pack_start(GTK_BOX(hbox), darea, FALSE, FALSE, 0);
	  gtk_widget_add_events(darea, GDK_LEAVE_NOTIFY_MASK | GDK_POINTER_MOTION_MASK );
	  gtk_signal_connect (GTK_OBJECT (darea), "expose-event",
			      GTK_SIGNAL_FUNC (on_darea_expose), NULL);
	  gtk_signal_connect (GTK_OBJECT(darea), "motion-notify-event",
			      GTK_SIGNAL_FUNC (on_mouse_move), NULL);
	  gtk_signal_connect (GTK_OBJECT(darea), "leave-notify-event",
			      GTK_SIGNAL_FUNC (on_focus_out), NULL);

	  vbox = gtk_vbox_new(FALSE, 0);
	  gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE, 0);

	  myframe = gtk_frame_new("Player Informations");
	  gtk_box_pack_start(GTK_BOX(vbox), myframe, TRUE, TRUE, 0);
	  
	  vbox2 = gtk_vbox_new(FALSE, 0);
	  gtk_container_add(GTK_CONTAINER(myframe), vbox2);

	  hbox2 = gtk_hbox_new(FALSE, 0);
	  gtk_box_pack_start(GTK_BOX(vbox2), hbox2, FALSE, FALSE, 0);

	  label = gtk_label_new("Id : ");
	  gtk_box_pack_start(GTK_BOX(hbox2), label, FALSE, FALSE, 0);
	  myid = gtk_label_new("N/A");
	  gtk_box_pack_start(GTK_BOX(hbox2), myid, TRUE, TRUE, 0);

	  label = gtk_label_new(" Money : ");
	  gtk_box_pack_start(GTK_BOX(hbox2), label, FALSE, FALSE, 0);
	  mymoney = gtk_label_new("N/A");
	  gtk_box_pack_start(GTK_BOX(hbox2), mymoney, TRUE, TRUE, 0);

	  label = gtk_label_new(" Capacity : ");
	  gtk_box_pack_start(GTK_BOX(hbox2), label, FALSE, FALSE, 0);
	  mycapacity = gtk_label_new("N/A");
	  gtk_box_pack_start(GTK_BOX(hbox2), mycapacity, TRUE, TRUE, 0);

	  sw = gtk_scrolled_window_new(NULL, NULL);
	  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), 
					 GTK_POLICY_AUTOMATIC, 
					 GTK_POLICY_AUTOMATIC);
	  gtk_box_pack_start(GTK_BOX(vbox2), sw, TRUE, TRUE, 0);
	  mypackets = gtk_clist_new_with_titles(3, mypacketstitles);
	  gtk_container_add(GTK_CONTAINER(sw), mypackets);

	  pframe = gtk_frame_new("All Known Players");
	  gtk_box_pack_start(GTK_BOX(vbox), pframe, FALSE, FALSE, 0);
	  pvbox = gtk_vbox_new(FALSE, 1);
	  gtk_container_add(GTK_CONTAINER(pframe), pvbox);

	  pkframe = gtk_frame_new("All known Packages");
	  gtk_box_pack_start(GTK_BOX(vbox), pkframe, TRUE, TRUE, 0);
	  sw = gtk_scrolled_window_new(NULL, NULL);
	  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), 
					 GTK_POLICY_AUTOMATIC, 
					 GTK_POLICY_AUTOMATIC);
	  gtk_container_add(GTK_CONTAINER(pkframe), sw);
	  packages = gtk_clist_new_with_titles(4, packagestitle);
	  gtk_container_add(GTK_CONTAINER(sw), packages);
	  
	  gsock = g_io_channel_unix_new(sock);
	  g_io_add_watch(gsock, G_IO_IN, on_data_in, NULL);
	  
	  gtk_widget_show_all(window);
	  gtk_main();
	  
	  close(sock);
	  return 0;
	} 
      }
    }
}
