#include "main.h"

#define FOUNDINBASE          1
#define ALLBASESEMTPY        0
#define SOMEBASESARENOTEMPTY 2

int get_prox_base(env *e)
{
  bedford_entry *be   = NULL;
  base *best_base     = NULL;
  base *b=e->bs->first;
  unsigned int min_path=((unsigned int)(-1));
  int ret = 0;
  
  be=bedford_lookup_target(e, e->p->x, e->p->y);
  while(b)
    {
      if(b->limit != (unsigned int)(-1))
	ret++;

      if(b->limit <= e->p->capacity)
	{
	  printf("Selection de %d, %d car limit = %u et cap = %u\n",
		 b->x, b->y, b->limit, e->p->capacity);
	  if( be->dist[b->x][b->y] < min_path )
	    {
	      best_base=b;
	      min_path=be->dist[b->x][b->y];
	    }
	} 
      else 
	printf("base %d, %d is removed\n", b->x, b->y);
      b=b->next;
    }
  
  if (best_base)
    {
      e->target = bedford_lookup_target(e,best_base->x,best_base->y);
      e->target_type=TBASE;
      return FOUNDINBASE;
    }
  if(ret > 0)
    return SOMEBASESARENOTEMPTY;
  else
    return ALLBASESEMTPY;
}

int select_nearest_target_by_pos(env *e, pkt_list *l)
{
  bedford_entry *be;
  pkt_entry   *best_pe   = NULL;
  unsigned int best_dist = (unsigned int)-1;
  pkt_list_entry *ple;

  if(!l || !l->first)
    return 0;

  be = bedford_lookup_target(e, e->p->x, e->p->y);

  for(ple = l->first; ple; ple = ple->next)
    if( be->dist[ple->entry->x][ple->entry->y] < best_dist )
      {
	best_dist = be->dist[ple->entry->x][ple->entry->y];
	best_pe   = ple->entry;
      }
  if(best_pe)
    {
      e->target = bedford_lookup_target(e, best_pe->x, best_pe->y);
      return 1;
    }
  return 0;
}

int select_nearest_target_by_home(env *e, pkt_list *l)
{
  bedford_entry *be;
  pkt_entry   *best_pe   = NULL;
  unsigned int best_dist = (unsigned int)-1;
  pkt_list_entry *ple;

  if(!l || !l->first)
    return 0;

  be = bedford_lookup_target(e, e->p->x, e->p->y);

  for(ple = l->first; ple; ple = ple->next)
    if( be->dist[ple->entry->xhome][ple->entry->yhome] < best_dist )
      {
	best_dist = be->dist[ple->entry->xhome][ple->entry->yhome];
	best_pe   = ple->entry;
      }
  if(best_pe)
    {
      e->target = bedford_lookup_target(e, best_pe->xhome, best_pe->yhome);
      return 1;
    }
  return 0;
}

int move_to_target(env *e)
{
  move(1, e, bestmove(e, e->target));
  return 0;
}

/*
int filled_enough(env *e)
{
  return 
    ((double)(e->p->packages.total_weight) 
     > COEF_FULL_BOUND * (double)(e->initial_capacity)) || 
    (allbases_limit(e->bs) < e->p->capacity);
}
*/

int filled_enough(env *e) { return 1; }

int player_reach_target(env *e)
{
  return e->target && (e->p->x == e->target->targetx) && (e->p->y == e->target->targety);
}

pkt_list * good_packages(env *e) 
{
  if ((e->p->available==NULL) || (e->p->available->first==NULL))
    return NULL;

  if (EMPTY_BAG(e))
    return binpack(e->p->available,e->initial_capacity,e);
  else
    return binpack_add(e->p->available,&e->p->packages,e->initial_capacity,e);
}



int get_dropped_packets(env *e)
{
  int res;
  pkt_list_entry *ple;
  pkt_list *sub=pkt_init();
  for(ple = e->known.first; ple; ple = ple->next)
    if(USABLE(ple->entry))
      pkt_add_entry(sub,ple->entry);
  res=select_nearest_target_by_pos(e,sub);
  e->target_type=TDP;
  pkt_free(sub);
  return res;
}

long int random(void);

void find_target_alea(env *e)
{
  bedford_entry *be;
  int x, y, r;
	
  be = bedford_lookup_target(e, e->p->x, e->p->y);
  r = (random() % 4) + 2;
  for( x = 0; x < e->b->width; x++)
    for( y = 0; y < e->b->height; y++)
      if( be->dist[x][y] == r ) {
	e->target = bedford_lookup_target(e, x, y);
	return;
      }
}

void find_target_for_pack(env *e)
{
  int ret;
  printf("Find Target for Package\n");
  ret = get_prox_base(e);
  if (ret == FOUNDINBASE)
    return;
  if (ret == ALLBASESEMTPY)
    {
      if (get_dropped_packets(e))
	return;
      find_target_alea(e);
      return;
    }
  assert(0); 
}

void find_target_for_dest(env *e)
{
  printf("Find Target for Dest\n");
  if(e->p->packages.first == NULL)
    {
      find_target_alea(e);
      return;
    }
  if (select_nearest_target_by_home(e, &e->p->packages))
    {
      printf("Selected Target : %d, %d\n", 
	     e->target->targetx, e->target->targety);
      e->target_type=TDEST;
      return;
    }
  assert(0);
}

void drop_local_packets(env *e)
{
  pkt_list_entry *ple;
  pkt_list *sub=pkt_init();
  for(ple = e->p->packages.first; ple; ple = ple->next)
    if ((ple->entry->xhome==e->p->x) && (ple->entry->yhome==e->p->y))
      pkt_add_entry(sub,ple->entry);
  assert(sub->first);
  drop(1,e,sub);
  pkt_free(sub);
}

int thinking(env *e)
{
  pkt_list_entry *ple;
  if( e->target_type==TDP )
    {
      for(ple = e->known.first; ple; ple = ple->next)
	if( (ple->entry->x == e->target->targetx) &&
	    (ple->entry->y == e->target->targety) && 
	    (!USABLE(ple->entry)) )
	  break;
      if(!ple) 
	{
	  printf("lost my target (tdp). Restarting from scratch\n");
	  e->state = EMPTY;
	}
    }
  
  if( e->target_type==TDEST ) {
    for(ple = e->p->packages.first; ple; ple = ple->next)
      if( (ple->entry->xhome == e->target->targetx) &&
	  (ple->entry->yhome == e->target->targety) )
	break;
    if(!ple)
      {
	if(!e->p->packages.first)
	  {
	    printf("lost all my packets and my target (TDEST). Restarting from scratch\n");
	    e->state = EMPTY;
	  }
	else
	  {
	    printf("lost all my target (TDEST) Changing it to first package's home\n");
	    e->target = bedford_lookup_target(e, e->p->packages.first->entry->xhome, 
					      e->p->packages.first->entry->yhome);
	  }
      }
  }
  return 1;
}


void decision(env *e)
{
  pkt_list *gp;

  (void)thinking(e);

  if(strafe_master(e) < 0)
    return;

  switch (e->state)
    {
    case EMPTY:
      printf("**EMPTY\n");
      gp=good_packages(e);
      if (gp)
	{
	  pick(1,e,gp);
	  e->state=PICKING;
	  return;
	}
      else
	{
	  printf("**EMPTY : no good packages, so find target for pack");
	  find_target_for_pack(e);
	  printf("gives %d, %d\n", e->target->targetx, e->target->targety);
	  move_to_target(e);
	  e->state=SEARCHING;
	  return;
	}
      break;
    case SEARCHING:
      printf("**SEARCHING\n");
      assert(e->target);
      gp=good_packages(e);
      if (gp)
	{
	  pick(1,e,gp);
	  e->state=PICKING;
	  return;
	}
      else
	{
	  if (player_reach_target(e))
	    {
	      if (EMPTY_PLACE(e))
		{
		  if (e->target_type==TBASE)
		    remove_base(e->bs,e->target->targetx,e->target->targety);
		  if (e->target_type==TDP)
		    remove_dropped_packet(e);
		  printf("**SEARCHING I am on an emply place, so I find target for pack ");
		  find_target_for_pack(e);
		  printf("gives %d, %d\n", e->target->targetx, e->target->targety);
		  move_to_target(e);
		  return;
		}
	      else
		{
		  if (filled_enough(e))
		    {
		      e->state=DELIVERING;
		      find_target_for_dest(e);
		      move_to_target(e);
		      return;
		    }
		  else
		    {
		      printf("**SEARCHING I am not filled enough, so I find target for pack ");
		      find_target_for_pack(e);
		      printf("gives %d, %d\n", e->target->targetx, e->target->targety);
		      move_to_target(e);
		      return;
		    }
		}
	    }
	  else
	    {
	      move_to_target(e);
	      return;
	    }
	}
      assert(0);/*not reachable*/
      break;
    case PICKING:
      printf("**PICKING\n");
      if (filled_enough(e))
	{
	  e->state=DELIVERING;
	  find_target_for_dest(e);
	  move_to_target(e);
	  return;
	}
      else
	{
	  e->state=SEARCHING;
	  printf("**PICKING I am not filled enough so I find target for pack ");
	  find_target_for_pack(e);
	  printf("gives %d, %d\n", e->target->targetx, e->target->targety);
	  move_to_target(e);
	  return;
	}
      assert(0);/*not reachable*/
      break;
    case DELIVERING:
      printf("**DELIVERING\n");
      if (EMPTY_BAG(e))
	{
	  printf("**DELIVERING I have an empty bag, so I find target for pack ");
	  find_target_for_pack(e);
	  printf("gives %d, %d\n", e->target->targetx, e->target->targety);
	  move_to_target(e);
	  e->state=SEARCHING;
	  return;
	}
      if (player_reach_target(e))
	{
	  drop_local_packets(e);
	  e->state=DROPPING;
	  return;
	}
      else
	{
	  move_to_target(e);
	  return;
	}
      assert(0);/*not reachable*/
      break;

    case DROPPING:
      printf("**DROPPING\n");
      if (EMPTY_BAG(e))
	{
	  printf("**DROPPING I have an empty bag, so I find target for pack ");
	  find_target_for_pack(e);
	  printf("gives %d %d\n", e->target->targetx, e->target->targety);
	  move_to_target(e);
	  e->state=SEARCHING;
	  return;
	}
      else
	{
	  e->state=DELIVERING;
	  find_target_for_dest(e);
	  move_to_target(e);
	  return;
	}
      assert(0);/*not reachable*/
      break;
    }
}


void decision2(env *e)
{
  pkt_list_entry *ple;
  if (!e->target)
    return;

  /*
   * Is target still not stupid ?
   */
    if( e->target_type==TDP )
      {
	for(ple = e->known.first; ple; ple = ple->next)
	  if( (ple->entry->x == e->target->targetx) &&
	      (ple->entry->y == e->target->targety) && (USABLE(ple->entry)))
	    break;
	if(!ple) 
	  {
	    printf("lost my target (tdp)\n");/*TODO*/
	  }
      }
  


  if( e->target_type==TDEST ) {
    for(ple = e->p->packages.first; ple; ple = ple->next)
      if( (ple->entry->xhome == e->target->targetx) &&
	  (ple->entry->yhome == e->target->targety) )
	break;
    if(!ple)
      {
	if(!e->p->packages.first)
	  {
	    printf("lost all my packets and my target (TDEST)\n");/*TODO*/
	  }
	else
	  {
	    printf("lost all my target (TDEST)\n");/*TODO*/
	  }
      }
  }

  if(strafe_master(e) < 0)
    return;

 
}
